

// DCP


// Includes
#include <ethernet.hpp>
#include <udp.hpp>


// Preprocessor Directives
#pragma once


// Check Compiler
#ifdef _MSC_BUILD

   // Undefine Macros
   #undef max
   #undef min

#endif


// Class DCP
class DCP
{
public:

   // Class Firmware
   class Code;

   // Class Config
   class Config;

   // Class Exception
   class Exception;

   // Class Info
   class Info;

   // Class Status
   class Status;

   // Message Types
   static const uint16_t MESSAGE_OK = 0x0001;
   static const uint16_t MESSAGE_ERROR = 0x0002;
   static const uint16_t MESSAGE_RESET = 0x0003;
   static const uint16_t MESSAGE_UPDATE = 0x0004;
   static const uint16_t MESSAGE_GET_INFO = 0x0005;
   static const uint16_t MESSAGE_INFO = 0x0006;
   static const uint16_t MESSAGE_GET_STATUS = 0x0007;
   static const uint16_t MESSAGE_STATUS = 0x0008;
   static const uint16_t MESSAGE_CONFIGURE = 0x0009;
   static const uint16_t MESSAGE_BURN = 0x000A;

   // Initialize
   static void init(void);

   // Constructor
   DCP(const IP::Address& ip_host = IP::Address(), uint16_t port_host = 0);

   // Destructor
   ~DCP(void);

   // Receive Packet with Timeout [ms]
   void receive(IP::Address& ip_source, uint16_t& message, uint32_t& id, std::string& str, uint16_t timeout = 0) const;

   // Send Packet
   void send(const IP::Address& ip_dest, uint16_t message, uint32_t id = 0, const std::string& str = "") const;

private:

   // Copy Constructor
   DCP(const DCP& dcp);

   // Class Header
   class _Header;

   // Class Option
   class _Option;

   // UDP Port
   static const uint16_t _PORT = 30000;

   // Variables
   UDP* _udp;

   // Friends
   friend std::string& operator <<(std::string& str, const _Header& header);
   friend std::string& operator >>(std::string& str, _Header& header);
   friend std::string& operator <<(std::string& str, const _Option& option);
   friend std::string& operator >>(std::string& str, _Option& option);
};


// Class Exception
class DCP::Exception
{
public:

   // Class Network
   class Network;

   // Class Parameter
   class Parameter;

   // Class Timeout
   class Timeout;

private:

   // Virtual Function for RTTI
   virtual void _func() {}
};


// Class Network
class DCP::Exception::Network : public DCP::Exception
{
};


// Class Parameter
class DCP::Exception::Parameter : public DCP::Exception
{
};


// Class Timeout
class DCP::Exception::Timeout : public DCP::Exception
{
};


// Class Code
class DCP::Code
{
public:

   // Constructor
   Code(uint16_t page, const std::string& data);

   // Convert to String
   operator const std::string(void) const;

   // Data
   const std::string& data(void) const;
   void data(const std::string& data);

   // Page
   uint16_t page(void) const;
   void page(uint16_t page);

private:

   // Variables
   uint16_t _page;
   std::string _data;
};


// Class Config
class DCP::Config
{
public:

   // DHCP Status
   static const uint8_t DHCP_DISABLED = 0;
   static const uint8_t DHCP_ENABLED = 1;

   // Constructor
   Config(void);
   Config(const std::string& name, uint8_t dhcp, const IP::Address& ip_host, const IP::Address& ip_subnet_mask,
      const IP::Address& ip_gateway);

   // Convert to String
   operator const std::string(void) const;

private:

   // Options
   static const uint8_t _OPTION_DHCP = 0x01;
   static const uint8_t _OPTION_IP_HOST = 0x02;
   static const uint8_t _OPTION_IP_SUBNET_MASK = 0x03;
   static const uint8_t _OPTION_IP_GATEWAY = 0x04;
   static const uint8_t _OPTION_NAME = 0x05;

   // Variables
   uint8_t _dhcp;
   std::string _name;
   IP::Address _ip_gateway;
   IP::Address _ip_host;
   IP::Address _ip_subnet_mask;
};


// Class Info
class DCP::Info
{
public:

   // Constructor
   Info(void);
   Info(const std::string& str);

   // Parse String
   const Info& operator =(const std::string& str);

   // Get Calibration Time
   const Time& calibration(void) const;

   // Get CPU Type
   const std::string& cpu(void) const;

   // Get Number of Power / Reset Cycles
   uint32_t cycles(void) const;

   // Get Device Description
   const std::string& description(void) const;

   // Get DHCP Status
   uint8_t dhcp(void) const;

   // Get Domain
   const std::string& domain(void) const;

   // Get Host Ethernet Address
   const Ethernet::Address& eth_host(void) const;

   // Get Expiration of Calibration Time
   const Time& expiration(void) const;

   // Get CPU Core Frequency [Hz]
   uint32_t frequency(void) const;

   // Get Hostname
   const std::string& hostname(void) const;

   // Get Host IP Address
   const IP::Address& ip_host(void) const;

   // Get Subnet Mask
   const IP::Address& ip_subnet_mask(void) const;

   // Get Gateway IP Address
   const IP::Address& ip_gateway(void) const;

   // Get Manufacturer
   const std::string& manufacturer(void) const;

   // Get external SRAM Memory Size [Bytes]
   uint32_t memory(void) const;

   // Get Device Model
   const std::string& model(void) const;

   // Get Name
   const std::string& name(void) const;

   // Get Network Adapter
   const std::string& network(void) const;

   // Get Firmware Release Time
   const Time& release(void) const;

   // Get Hardware Revision
   const std::string& revision(void) const;

   // Get Serial Number
   uint32_t serial(void) const;

   // Get Tag
   const std::string& tag(void) const;

   // Get Uptime since Reset [s]
   uint32_t uptime_reset(void) const;

   // Get total Uptime [s]
   uint32_t uptime_total(void) const;

   // Get Firmware Version
   const std::string& version(void) const;

private:

   // Options
   static const uint8_t _OPTION_MODEL = 0x01;
   static const uint8_t _OPTION_DESCRIPTION = 0x02;
   static const uint8_t _OPTION_MANUFACTURER = 0x03;
   static const uint8_t _OPTION_SERIAL = 0x04;
   static const uint8_t _OPTION_REVISION = 0x05;
   static const uint8_t _OPTION_CPU = 0x06;
   static const uint8_t _OPTION_NETWORK = 0x07;
   static const uint8_t _OPTION_FREQUENCY = 0x08;
   static const uint8_t _OPTION_MEMORY = 0x09;
   static const uint8_t _OPTION_VERSION = 0x0A;
   static const uint8_t _OPTION_RELEASE = 0x0B;
   static const uint8_t _OPTION_CALIBRATION = 0x0C;
   static const uint8_t _OPTION_EXPIRATION = 0x0D;
   static const uint8_t _OPTION_UPTIME_RESET = 0x0E;
   static const uint8_t _OPTION_UPTIME_TOTAL = 0x0F;
   static const uint8_t _OPTION_CYCLES = 0x10;
   static const uint8_t _OPTION_ETH_HOST = 0x11;
   static const uint8_t _OPTION_IP_HOST = 0x12;
   static const uint8_t _OPTION_IP_SUBNET_MASK = 0x13;
   static const uint8_t _OPTION_IP_GATEWAY = 0x14;
   static const uint8_t _OPTION_HOSTNAME = 0x15;
   static const uint8_t _OPTION_DOMAIN = 0x16;
   static const uint8_t _OPTION_DHCP = 0x17;
   static const uint8_t _OPTION_NAME = 0x18;
   static const uint8_t _OPTION_TAG = 0x19;

   // Variables
   uint8_t _dhcp;
   uint32_t _cycles;
   uint32_t _frequency;
   uint32_t _memory;
   uint32_t _serial;
   uint32_t _uptime_reset;
   uint32_t _uptime_total;
   std::string _cpu;
   std::string _description;
   std::string _domain;
   std::string _hostname;
   std::string _model;
   std::string _manufacturer;
   std::string _name;
   std::string _network;
   std::string _revision;
   std::string _tag;
   std::string _version;
   Time _calibration;
   Time _expiration;
   Time _release;
   Ethernet::Address _eth_host;
   IP::Address _ip_gateway;
   IP::Address _ip_host;
   IP::Address _ip_subnet_mask;
};


// Class Status
class DCP::Status
{
public:

   // Class Property
   class Property;

   // Constructor
   Status(void);
   Status(const std::string& str);

   // Parse String
   const Status& operator =(const std::string& str);

   // Check if OK
   bool ok(void) const;

   // Get Property
   const Property& operator [](const std::string& symbol) const;
   const std::map<std::string, Property>& property(void) const;
   const Property& property(const std::string& symbol) const;

private:

   // Options
   static const uint8_t _OPTION_SYMBOL = 0x01;
   static const uint8_t _OPTION_DESCRIPTION = 0x02;
   static const uint8_t _OPTION_UNIT = 0x03;
   static const uint8_t _OPTION_TYPICAL = 0x04;
   static const uint8_t _OPTION_MIN = 0x05;
   static const uint8_t _OPTION_MAX = 0x06;
   static const uint8_t _OPTION_VALUE = 0x07;

   // Variables
   bool _ok;
   std::map<std::string, Property> _property;
};


// Class Property
class DCP::Status::Property
{
public:

   // Get Description
   const std::string& description(void) const;

   // Get minimum Value
   float min(void) const;

   // Get maximum Value
   float max(void) const;

   // Check if OK
   bool ok(void) const;

   // Get Symbol
   const std::string& symbol(void) const;

   // Get typical Value
   float typical(void) const;

   // Get Unit
   const std::string& unit(void) const;

   // Get Value
   float value(void) const;

private:

   // Constructor
   Property(void);

   // Variables
   bool _ok;
   float _min;
   float _max;
   float _typical;
   float _value;
   std::string _description;
   std::string _symbol;
   std::string _unit;

   // Friends
   friend Status;
};
