

// DEVICE


// Includes
#include <vector>
#include <time.hpp>
#include "dcp.hpp"


// Preprocessor Directives
#pragma once


// Class Device
class Device
{
public:

   // Class Exception
   class Exception;

   // Class Firmware
   class Firmware;

   // Class Info
   class Info;

   // Class Status
   class Status;

   // Firmware Update Mode
   static const uint8_t UPDATE_COMATIBLE = 1;
   static const uint8_t UPDATE_UNCONDITIONAL = 2;

   // Configure
   static void configure(uint32_t serial);
   static void configure(uint32_t serial, const std::string& name, uint8_t dhcp, const IP::Address& ip_host,
      const IP::Address& ip_subnet_mask, const IP::Address& ip_gateway);

   // Find Devices in Subnet
   static std::set<Device> find(void);
   static Device find(uint32_t serial);
   static std::map<uint32_t, Device> find(std::set<uint32_t> serial);

   // Initialize
   static void init(void);

   // Constructor
   Device(const IP::Address& address);

   // IP Address
   const IP::Address& address(void) const;
   void address(const IP::Address& address);

   // Burn Tag
   void burn(const std::string& tag);

   // Get Info
   Info info(void) const;

   // Reset
   void reset(void) const;

   // Get Status
   Status status(void) const;

   // Timeout [ms]
   uint16_t timeout(void) const;
   void timeout(uint16_t timeout);

   // Update Firmware
   void update(const Firmware& firmware, uint8_t update = UPDATE_COMATIBLE) const;

protected:

   // Class DCP
   class _DCP;

   // Request
   uint16_t _request(uint16_t message, std::string& str) const;

   // Get unique ID
   static uint32_t _unique_id(void);

   // Variables
   IP::Address _address;

private:

   // Class Broadcast
   class _Broadcast;

   // Class Option
   class _Option;

   // Default Timeout [ms]
   static const uint16_t _TIMEOUT = 1000;

   // Configure
   static void _configure(uint32_t serial, const DCP::Config& config = DCP::Config());

   // Operator <
   bool operator <(const Device& device) const;

   // Variables
   static std::map<uint32_t, uint16_t> _timeout;

   // Friends
   friend std::less<Device>;
   friend std::string& operator <<(std::string& str, const _Option& option);
   friend std::string& operator >>(std::string& str, _Option& option);
};


// Class Exception
class Device::Exception
{
public:

   // Class Failed
   class Failed;

   // Class Network
   class Network;

   // Class NotFound
   class NotFound;

   // Class Parameter
   class Parameter;

   // Class Timeout
   class Timeout;

private:

   // Virtual Function for RTTI
   virtual void _func() {}
};


// Class Failed
class Device::Exception::Failed : public Device::Exception
{
};


// Class Network
class Device::Exception::Network : public Device::Exception
{
};


// Class NotFound
class Device::Exception::NotFound : public Device::Exception
{
};


// Class Parameter
class Device::Exception::Parameter : public Device::Exception
{
};


// Class Timeout
class Device::Exception::Timeout : public Device::Exception
{
};


// Class Firmware
class Device::Firmware
{
public:

   // Constructor
   Firmware(void);
   Firmware(const std::string& str);

   // Parse String
   const Firmware& operator =(const std::string& str);

   // Convert to String
   operator std::string(void) const;

   // Code
   const std::string code(void) const;
   void code(const std::string& code);

   // CPU Type
   const std::string& cpu(void) const;
   void cpu(const std::string& cpu);

   // CPU Core Frequency [Hz]
   uint32_t frequency(void) const;
   void frequency(uint32_t frequency);

   // Manufacturer
   const std::string& manufacturer(void) const;
   void manufacturer(const std::string& manufacturer);

   // External SRAM Memory Size [Bytes]
   uint32_t memory(void) const;
   void memory(uint32_t memory);

   // Device Model
   const std::string& model(void) const;
   void model(const std::string& model);

   // Network Adapter
   const std::string& network(void) const;
   void network(const std::string& network);

   // Firmware Release Time
   const Time& release(void) const;
   void release(const Time& release);

   // Hardware Revision
   const std::string& revision(void) const;
   void revision(const std::string& revision);

   // Firmware Version
   const std::string& version(void) const;
   void version(const std::string& version);

private:

   // Options
   static const uint8_t _OPTION_MODEL = 0x01;
   static const uint8_t _OPTION_MANUFACTURER = 0x02;
   static const uint8_t _OPTION_REVISION = 0x03;
   static const uint8_t _OPTION_CPU = 0x04;
   static const uint8_t _OPTION_FREQUENCY = 0x05;
   static const uint8_t _OPTION_MEMORY = 0x06;
   static const uint8_t _OPTION_NETWORK = 0x07;
   static const uint8_t _OPTION_VERSION = 0x08;
   static const uint8_t _OPTION_RELEASE = 0x09;

   // Variables
   uint32_t _frequency;
   uint32_t _memory;
   std::string _code;
   std::string _cpu;
   std::string _manufacturer;
   std::string _model;
   std::string _network;
   std::string _revision;
   std::string _version;
   Time _release;
};


// Class Info
class Device::Info : public DCP::Info
{
public:

   // Constructor
   Info(void);

private:

   // Constructor
   Info(const std::string& str);

   // Parse String
   using DCP::Info::operator =;

   // Friends
   friend Device;
};


// Class Status
class Device::Status : public DCP::Status
{
public:

   // Constructor
   Status(void);

private:

   // Constructor
   Status(const std::string& str);

   // Parse String
   using DCP::Status::operator =;

   // Friends
   friend Device;
};


// Class DCP
class Device::_DCP
{
public:

   // Constructor
   _DCP(void);

   // Destructor
   ~_DCP(void);

   // Get DCP Socket
   DCP* operator ->(void);

private:

   // Variables
   static std::vector<std::pair<DCP*, bool>> _pool;
   std::vector<std::pair<DCP*, bool>>::iterator _dcp;
};
