

// SPLINE 3.1.0


// Includes
#include <vector>
#include <math.h>


// Check Compiler
#ifdef _MSC_BUILD

   // Undefine Macros
   #undef min
   #undef max

#else

   // Includes
   #include <stddef.h>

#endif


// Preprocessor Directives
#pragma once


// Class Spline
template <typename T> class Spline
{
public:

   // Class Exception
   class Exception;

   // Class Point
   class Point;

   // Swap Coordinates
   static const uint8_t SWAP_NO = 0;
   static const uint8_t SWAP_XY = 1;

   // Constructor
   Spline(void);
   Spline(const Point* X, size_t n, uint8_t swap = SWAP_NO);

   // Evaluate
   const T operator ()(const T& x) const;

   // Get lower Limit
   const T& min(void) const;

   // Get upper Limit
   const T& max(void) const;

private:

   // Variables
   std::vector<std::vector<T>> _a;
};


// Class Exception
template <typename T> class Spline<T>::Exception
{
public:

   // Class Parameter
   class Parameter;

private:

   // Virtual Function for RTTI
   virtual void _func() {}
};


// Class Parameter
template <typename T> class Spline<T>::Exception::Parameter : public Spline::Exception
{
};


// Class Point
template <typename T> class Spline<T>::Point
{
public:

   // Variables
   T x;
   T y;
};


// Swap Coordinates
template <typename T> const uint8_t Spline<T>::SWAP_NO;
template <typename T> const uint8_t Spline<T>::SWAP_XY;


// Constructor
template <typename T> inline Spline<T>::Spline(void) : _a(std::vector<std::vector<T>>(1, std::vector<T>(5, NAN)))
{
}


// Constructor
template <typename T> Spline<T>::Spline(const Point* X, size_t n, uint8_t swap)
{
   // Check Number of Data Points
   if (n < 3)
   {
      // Exception
      throw typename Exception::Parameter();
   }

   // Create auxiliary Variables
   std::vector<T> h(n - 1);
   std::vector<T> b(n - 1);
   std::vector<T> u(n - 2);
   std::vector<T> v(n - 2);
   std::vector<T> z(n);

   // Parse Data Points
   for (size_t i = 0; i < (n - 1); ++i)
   {
      // Compute h[i]
      h[i] = (swap == SWAP_XY) ? (X[i + 1].y - X[i].y) : (X[i + 1].x - X[i].x);

      // Check Difference
      if (h[i] <= 0)
      {
         // Exception
         throw typename Exception::Parameter();
      }

      // Compute b[i]
      b[i] = ((swap == SWAP_XY) ? (X[i + 1].x - X[i].x) : (X[i + 1].y - X[i].y)) / h[i];
   }

   // Set Constraints
   u[0] = 2 * (h[0] + h[1]);
   v[0] = 6 * (b[1] - b[0]);

   // Parse Data Points
   for (size_t i = 1; i < (n - 2); ++i)
   {
      // Perform Gaussian Elimination
      u[i] = 2 * (h[i + 1] + h[i]) - h[i] * h[i] / u[i - 1];
      v[i] = 6 * (b[i + 1] - b[i]) - h[i] * v[i - 1] / u[i - 1];
   }

   // Set Constraints
   z[0] = 0;
   z[n - 1] = 0;

   // Perform back-Substitution
   for (size_t i = (n - 2); 0 < i; --i)
   {
      z[i] = (v[i - 1] - h[i] * z[i + 1]) / u[i - 1];
   }

   // Create Coefficients
   _a = std::vector<std::vector<T>>(n - 1, std::vector<T>(5));

   // Compute Coefficients and Domain
   for (size_t i = 0; i < (n - 1); ++i)
   {
      _a[i][0] = (swap == SWAP_XY) ? X[i].y : X[i].x;
      _a[i][1] = (swap == SWAP_XY) ? X[i].x : X[i].y;
      _a[i][2] = -h[i] * z[i + 1] / 6 - h[i] * z[i] / 3 + b[i]; // (X[i + 1].y - X[i].y) / h[i]  instead of b[i] ***
      _a[i][3] = z[i] / 2;
      _a[i][4] = (z[i + 1] - z[i]) / 6 / h[i];
   }
}


// Evaluate
template <typename T> const T Spline<T>::operator ()(const T& x) const
{
   // Indices
   size_t i = 0;
   size_t j = static_cast<size_t>(_a.size());
   size_t k;

   // Find Polynom (binary Search)
   while ((i + 1) < j)
   {
      // Split Range
      k = (i + j) / 2;

      // Check Range
      if (_a[k][0] < x)
      {
         // Discard lower Half
         i = k;
      }
      else
      {
         // Discard upper Half
         j = k;
      }
   }

   // Evaluate Polynomial
   return _a[i][1] + (x - _a[i][0]) * (_a[i][2] + (x - _a[i][0]) * (_a[i][3] + (x - _a[i][0]) * _a[i][4]));
}


// Get lower Limit
template <typename T> inline const T& Spline<T>::min(void) const
{
   // Return lower Limit
   return _a[0][0];
}


// Get upper Limit
template <typename T> inline const T& Spline<T>::max(void) const
{
   // Return upper Limit
   return _a[_a.size() - 1][0];
}
