

// TMP01 - OCTAL THERMOCOUPLE MONITOR


// Includes
#include <spline.hpp>
#include "device.hpp"


// Preprocessor Directives
#pragma once


// Class TMP01
class TMP01 : public Device
{
public:

   // Class Calibration
   class Calibration;

   // Class Exception
   class Exception;

   // Class Sensor
   class Sensor;

   // Channels
   static const uint8_t CHANNELS = 8;

   // Sampling Frequencies [Hz]
   static const uint16_t FREQUENCY_6 = 6;
   static const uint16_t FREQUENCY_12 = 12;
   static const uint16_t FREQUENCY_25 = 25;
   static const uint16_t FREQUENCY_50 = 50;
   static const uint16_t FREQUENCY_100 = 100;
   static const uint16_t FREQUENCY_200 = 200;
   static const uint16_t FREQUENCY_400 = 400;
   static const uint16_t FREQUENCY_800 = 800;
   static const uint16_t FREQUENCY_1500 = 1500;
   static const uint16_t FREQUENCY_3000 = 3000;

   // Constructor
   TMP01(const IP::Address& address);
   TMP01(const Device& device);

   // Calibrate
   void calibrate(void) const;
   void calibrate(const Calibration& calibration) const;

   // Set Sampling Frequency [Hz]
   void frequency(uint16_t frequency) const;

   // Measure Temperature [K]
   float measure(uint8_t channel) const;
   std::map<uint8_t, float> measure(void) const;
   std::map<uint8_t, float> measure(const std::set<uint8_t>& channel) const;

   // Set Sensor
   void sensor(const Sensor& sensor) const;
   void sensor(uint8_t channel, const Sensor& sensor) const;
   void sensor(const std::set<uint8_t>& channel, const Sensor& sensor) const;
   void sensor(const std::map<uint8_t, Sensor>& sensor) const;

   // Stop Calibration
   void stop(void) const;

private:

   // DCP Message Types
   static const uint16_t _MESSAGE_MEASURE = 0x0600;
   static const uint16_t _MESSAGE_TEMPERATURE = 0x0601;
   static const uint16_t _MESSAGE_ERROR_CONFIGURATION = 0x0602;
   static const uint16_t _MESSAGE_SET_CALIBRATION = 0x0603;
   static const uint16_t _MESSAGE_BUSY = 0x0604;
   static const uint16_t _MESSAGE_ERROR_RESERVED = 0x0605;
   static const uint16_t _MESSAGE_ERROR_ABORTED = 0x0606;
   static const uint16_t _MESSAGE_STOP = 0x0607;
   static const uint16_t _MESSAGE_SET_FREQUENCY = 0x0608;
   static const uint16_t _MESSAGE_SET_SENSOR = 0x0609;

   // Calibration Expiration [Days]
   static const uint16_t _CALIBRATION_EXPIRATION = 365;
   
   // Cold-Junction Compensation Correction [K]
   static const float _CJC_CORRECTION;

   // Sensor Curves
   static const Spline<float>::Point _CURVE_PT[];
   static const Spline<float> _curve_pt;
};


// Class Calibration
class TMP01::Calibration
{
public:

   // Constructor
   Calibration(void);
   Calibration(const std::string& str);

   // Convert to String
   operator std::string(void) const;

   // Parse String
   Calibration& operator =(const std::string& str);

   // Voltage Coefficient List
   const std::vector<float>& voltage(void) const;
   void voltage(const std::vector<float>& coeff);

private:

   // Polynomial Degree
   static const uint8_t _DEGREE = 7;

   // Variables
   std::vector<float> _voltage;
};


// Class Exception
class TMP01::Exception : public Device::Exception
{
public:

   // Class Aborted
   class Aborted;

   // Class Configuration
   class Configuration;

   // Class Reserved
   class Reserved;
};


// Class Aborted
class TMP01::Exception::Aborted : public TMP01::Exception
{
};


// Class Configuration
class TMP01::Exception::Configuration : public TMP01::Exception
{
};


// Class Reserved
class TMP01::Exception::Reserved : public TMP01::Exception
{
};


// Class Sensor
class TMP01::Sensor
{
public:

   // Types
   static const uint8_t TYPE_NONE = 0;
   static const uint8_t TYPE_CUSTOM = 1;
   static const uint8_t TYPE_B = 2;
   static const uint8_t TYPE_C = 3;
   static const uint8_t TYPE_D = 4;
   static const uint8_t TYPE_E = 5;
   static const uint8_t TYPE_G = 6;
   static const uint8_t TYPE_J = 7;
   static const uint8_t TYPE_K = 8;
   static const uint8_t TYPE_M = 9;
   static const uint8_t TYPE_N = 10;
   static const uint8_t TYPE_P = 11;
   static const uint8_t TYPE_R = 12;
   static const uint8_t TYPE_S = 13;
   static const uint8_t TYPE_T = 14;

   // Constructor
   Sensor(uint8_t type = TYPE_NONE);

   // Type
   uint8_t type(void) const;
   void type(uint8_t type);

private:

   // Sensor Curves
   static const Spline<float>::Point _CURVE_B[];
   static const Spline<float>::Point _CURVE_C[];
   static const Spline<float>::Point _CURVE_D[];
   static const Spline<float>::Point _CURVE_E[];
   static const Spline<float>::Point _CURVE_G[];
   static const Spline<float>::Point _CURVE_J[];
   static const Spline<float>::Point _CURVE_K[];
   static const Spline<float>::Point _CURVE_M[];
   static const Spline<float>::Point _CURVE_N[];
   static const Spline<float>::Point _CURVE_P[];
   static const Spline<float>::Point _CURVE_R[];
   static const Spline<float>::Point _CURVE_S[];
   static const Spline<float>::Point _CURVE_T[];
   
   // Initialize
   static void _init(void);
   
   // Convert to String
   operator std::string(void) const;

   // Evaluate
   float _eval(float voltage, float temperature) const;

   // Variables
   static std::map<uint8_t, Spline<float>> _curve_temperature;
   static std::map<uint8_t, Spline<float>> _curve_voltage;
   uint8_t _type;

   // Friends
   friend TMP01;
};
