

// TMP02 - QUAD RTD MONITOR


// Includes
#include <spline.hpp>
#include "device.hpp"


// Preprocessor Directives
#pragma once


// Class TMP02
class TMP02 : public Device
{
public:

   // Class Calibration
   class Calibration;

   // Class Exception
   class Exception;

   // Class Sensor
   class Sensor;

   // Channels
   static const uint8_t CHANNELS = 4;

   // Sampling Frequencies [Hz]
   static const uint16_t FREQUENCY_6 = 6;
   static const uint16_t FREQUENCY_12 = 12;
   static const uint16_t FREQUENCY_25 = 25;
   static const uint16_t FREQUENCY_50 = 50;
   static const uint16_t FREQUENCY_100 = 100;
   static const uint16_t FREQUENCY_200 = 200;
   static const uint16_t FREQUENCY_400 = 400;
   static const uint16_t FREQUENCY_800 = 800;
   static const uint16_t FREQUENCY_1500 = 1500;
   static const uint16_t FREQUENCY_3000 = 3000;

   // Constructor
   TMP02(const IP::Address& address);
   TMP02(const Device& device);

   // Calibrate
   void calibrate(void) const;
   void calibrate(const Calibration& calibration) const;

   // Set Sampling Frequency [Hz]
   void frequency(uint16_t frequency) const;

   // Measure Temperature [K]
   float measure(uint8_t channel) const;
   std::map<uint8_t, float> measure(void) const;
   std::map<uint8_t, float> measure(const std::set<uint8_t>& channel) const;

   // Set Sensor Type
   void sensor(const Sensor& sensor) const;
   void sensor(uint8_t channel, const Sensor& sensor) const;
   void sensor(const std::set<uint8_t>& channel, const Sensor& sensor) const;
   void sensor(const std::map<uint8_t, Sensor>& sensor) const;

   // Stop Calibration
   void stop(void) const;

private:

   // DCP Message Types
   static const uint16_t _MESSAGE_MEASURE = 0x0200;
   static const uint16_t _MESSAGE_TEMPERATURE = 0x0201;
   static const uint16_t _MESSAGE_ERROR_CONFIGURATION = 0x0202;
   static const uint16_t _MESSAGE_SET_CALIBRATION = 0x0203;
   static const uint16_t _MESSAGE_BUSY = 0x0204;
   static const uint16_t _MESSAGE_ERROR_RESERVED = 0x0205;
   static const uint16_t _MESSAGE_ERROR_ABORTED = 0x0206;
   static const uint16_t _MESSAGE_STOP = 0x0207;
   static const uint16_t _MESSAGE_SET_FREQUENCY = 0x0208;
   static const uint16_t _MESSAGE_SET_SENSOR = 0x0209;

   // Calibration Expiration [Days]
   static const uint16_t _CALIBRATION_EXPIRATION = 365;
};


// Class Calibration
class TMP02::Calibration
{
public:

   // Excitation Currents [A]
   static const float CURRENT_10UA;
   static const float CURRENT_1MA;

   // Voltage Ranges [V]
   static const float VOLTAGE_0V5;
   static const float VOLTAGE_1V;
   static const float VOLTAGE_2V5;
   static const float VOLTAGE_5V;

   // Constructor
   Calibration(void);
   Calibration(const std::string& str);

   // Convert to String
   operator std::string(void) const;

   // Parse String
   Calibration& operator =(const std::string& str);

   // Excitation Current Coefficient
   float current(float current) const;
   void current(float current, float coeff);

   // Voltage Coefficient List
   const std::vector<float>& voltage(float voltage) const;
   void voltage(float voltage, const std::vector<float>& coeff);

private:

   // Polynomial Degree
   static const uint8_t _DEGREE = 7;

   // Initialize
   void _init(void);

   // Variables
   std::map<float, float> _current;
   std::map<float, std::vector<float>> _voltage;
};


// Class Exception
class TMP02::Exception : public Device::Exception
{
public:

   // Class Aborted
   class Aborted;

   // Class Configuration
   class Configuration;

   // Class Reserved
   class Reserved;
};


// Class Aborted
class TMP02::Exception::Aborted : public TMP02::Exception
{
};


// Class Configuration
class TMP02::Exception::Configuration : public TMP02::Exception
{
};


// Class Reserved
class TMP02::Exception::Reserved : public TMP02::Exception
{
};


// Class Sensor
class TMP02::Sensor
{
public:

   // Excitation Currents [A]
   static const float CURRENT_DEFAULT;
   static const float CURRENT_NONE;
   static const float CURRENT_10UA;
   static const float CURRENT_1MA;

   // Nominal Resistance Limits [Ohms]
   static const float RESISTANCE_DEFAULT;
   static const float RESISTANCE_MIN;
   static const float RESISTANCE_MAX;

   // Types
   static const uint8_t TYPE_NONE = 0;
   static const uint8_t TYPE_CUSTOM = 1;
   static const uint8_t TYPE_PT = 2;
   static const uint8_t TYPE_DT470 = 3;
   static const uint8_t TYPE_DT670 = 4;
   static const uint8_t TYPE_KTY81_1 = 5;
   static const uint8_t TYPE_KTY81_2 = 6;
   static const uint8_t TYPE_KTY82_1 = 7;
   static const uint8_t TYPE_KTY82_2 = 8;
   static const uint8_t TYPE_KTY83_1 = 9;
   static const uint8_t TYPE_KTY84_1 = 10;

   // Voltage Ranges [V]
   static const float VOLTAGE_DEFAULT;
   static const float VOLTAGE_0V5;
   static const float VOLTAGE_1V;
   static const float VOLTAGE_2V5;
   static const float VOLTAGE_5V;

   // Constructor
   Sensor(uint8_t type = TYPE_NONE);
   Sensor(uint8_t type, float voltage, float current);
   Sensor(uint8_t type, float resistance);

   // Excitation Current [A]
   float current(void) const;
   void current(float current);

   // Nominal Resistance [Ohms]
   float resistance(void) const;
   void resistance(float resistance);

   // Type
   uint8_t type(void) const;
   void type(uint8_t type);

   // Voltage Range [V]
   float voltage(void) const;
   void voltage(float voltage);

private:

   // Sensor Curves
   static const Spline<float>::Point _CURVE_PT[];
   static const Spline<float>::Point _CURVE_DT470[];
   static const Spline<float>::Point _CURVE_DT670[];
   static const Spline<float>::Point _CURVE_KTY81_82_1[];
   static const Spline<float>::Point _CURVE_KTY81_82_2[];
   static const Spline<float>::Point _CURVE_KTY83_1[];
   static const Spline<float>::Point _CURVE_KTY84_1[];

   // Initialize
   static void _init(void);

   // Convert to String
   operator std::string(void) const;

   // Evaluate
   float _eval(float voltage) const;

   // Variables
   static std::map<uint8_t, Spline<float>> _curve;
   uint8_t _type;
   float _current;
   float _resistance;
   float _voltage;

   // Friends
   friend TMP02;
};
